/**************************************
****   DEBUG.C  - Debug Commands   ****
**************************************/

//-- Include Files -----------------------------------------------------------
#include <stdio.h>
#include <conio.h>
#include "../memory/memory.h"
#include "../mc68000/68kregs.h"
#include "../video/video.h"

//-- Imported Variables ------------------------------------------------------
extern	Registers68K	regs;

//-- Imported Functions ------------------------------------------------------
extern	int				m68k_disassemble(char*, int);

//-- Defines -----------------------------------------------------------------
#define min(a,b)	((a) < (b) ? (a) : (b))

//-- Private Variables -------------------------------------------------------
static char	dasm_buffer[256];

//-- Function Prototypes -----------------------------------------------------
void	debug_dump_regs(void);
void	debug_dump_dasm_pc(void);
int		debug_dump_mem(int, int);
int		debug_dump_vidmem(int Addr, int Size);

//----------------------------------------------------------------------------
void	debug_dump_regs(void)
{
printf("D0:%08x D1:%08x D2:%08x D3:%08x\nD4:%08x D5:%08x D6:%08x D7:%08x\n", 
	regs.d0, regs.d1, regs.d2, regs.d3, regs.d4, regs.d5, regs.d6, regs.d7);
printf("A0:%08x A1:%08x A2:%08x A3:%08x\nA4:%08x A5:%08x A6:%08x A7:%08x\n", 
	regs.a0, regs.a1, regs.a2, regs.a3, regs.a4, regs.a5, regs.a6, regs.a7);
	printf("USP:%08x ISP:%08x SR:%02x%02x PC:%08x\n", regs.usp, regs.isp, 
		regs.sr_h&0xFF, regs.ccr&0xFF, regs.pc);
}

//----------------------------------------------------------------------------
void	debug_dump_dasm_pc(void)
{
	m68k_disassemble(dasm_buffer, regs.pc);
	printf("%08x: %s\n", regs.pc, dasm_buffer);
}

//----------------------------------------------------------------------------
int	debug_dump_mem(int Addr, int Size)
{
	int	i, val;
	int	x, y;
	
	while(Size>0) {
		printf("%08x: ", Addr);
		fflush(stdout);
		x = wherex();
		y = wherey();
		for(i=0;i<min(16, Size);i++) {
			val = cpu_readmem24(Addr++);
			gotoxy(x+i*3,y);
			printf("%02x ", (unsigned char)val);
			fflush(stdout);
			if (val<32)
				val = 0x2E;
			gotoxy(x+48+i, y);
			printf("%c", val);
			fflush(stdout);
		}
		printf("\n");
		Size -= 16;
	}
	
	return	Addr;
}

//----------------------------------------------------------------------------
int	debug_dump_vidmem(int Addr, int Size)
{
	int	i, val;
	int	x, y;
	
	while(Size>0) {
		Addr &= 0x1FFFF;
		printf("%05x: ", Addr);
		fflush(stdout);
		x = wherex();
		y = wherey();
		for(i=0;i<min(16, Size);i++) {
			val = *(video_vidram + (Addr ^ 1));
			Addr++;
			gotoxy(x+i*3,y);
			printf("%02x ", (unsigned char)val);
			fflush(stdout);
			if (val<32)
				val = 0x2E;
			gotoxy(x+48+i, y);
			printf("%c", val);
			fflush(stdout);
		}
		printf("\n");
		Size -= 16;
	}
	
	return	Addr;
}
