/**************************************
****   INPUT.C  -  Input devices   ****
**************************************/

//-- Include Files -----------------------------------------------------------
#include <stdio.h>
#include <allegro.h>
#include "../config/config.h"
#include "macro.h"

//-- Structures --------------------------------------------------------------
typedef struct {
	unsigned char	key_name[16];
	unsigned int	key_scancode;
} KEYDEF;

//-- Private Variables -------------------------------------------------------
static KEYDEF		keydefs[95] = {
	{"KEY_ESC",                1},
	{"KEY_1",                  2},
	{"KEY_2",                  3},
	{"KEY_3",                  4},
	{"KEY_4",                  5},
	{"KEY_5",                  6},
	{"KEY_6",                  7},
	{"KEY_7",                  8},
	{"KEY_8",                  9},
	{"KEY_9",                 10},
	{"KEY_0",                 11},
	{"KEY_MINUS",             12},
	{"KEY_EQUALS",            13},
	{"KEY_BACKSPACE",         14},
	{"KEY_TAB",               15},
	{"KEY_Q",                 16},
	{"KEY_W",                 17},
	{"KEY_E",                 18},
	{"KEY_R",                 19},
	{"KEY_T",                 20},
	{"KEY_Y",                 21},
	{"KEY_U",                 22},
	{"KEY_I",                 23},
	{"KEY_O",                 24},
	{"KEY_P",                 25},
	{"KEY_OPENBRACE",         26},
	{"KEY_CLOSEBRACE",        27},
	{"KEY_ENTER",             28},
	{"KEY_CONTROL",           29},
	{"KEY_LCONTROL",          29},
	{"KEY_A",                 30},
	{"KEY_S",                 31},
	{"KEY_D",                 32},
	{"KEY_F",                 33},
	{"KEY_G",                 34},
	{"KEY_H",                 35},
	{"KEY_J",                 36},
	{"KEY_K",                 37},
	{"KEY_L",                 38},
	{"KEY_COLON",             39},
	{"KEY_QUOTE",             40},
	{"KEY_TILDE",             41},
	{"KEY_LSHIFT",            42},
	{"KEY_BACKSLASH",         43},
	{"KEY_Z",                 44},
	{"KEY_X",                 45},
	{"KEY_C",                 46},
	{"KEY_V",                 47},
	{"KEY_B",                 48},
	{"KEY_N",                 49},
	{"KEY_M",                 50},
	{"KEY_COMMA",             51},
	{"KEY_STOP",              52},
	{"KEY_SLASH",             53},
	{"KEY_RSHIFT",            54},
	{"KEY_ASTERISK",          55},
	{"KEY_ALT",               56},
	{"KEY_SPACE",             57},
	{"KEY_CAPSLOCK",          58},
	{"KEY_F1",                59},
	{"KEY_F2",                60},
	{"KEY_F3",                61},
	{"KEY_F4",                62},
	{"KEY_F5",                63},
	{"KEY_F6",                64},
	{"KEY_F7",                65},
	{"KEY_F8",                66},
	{"KEY_F9",                67},
	{"KEY_F10",               68},
	{"KEY_NUMLOCK",           69},
	{"KEY_SCRLOCK",           70},
	{"KEY_HOME",              71},
	{"KEY_UP",                72},
	{"KEY_PGUP",              73},
	{"KEY_MINUS_PAD",         74},
	{"KEY_LEFT",              75},
	{"KEY_5_PAD",             76},
	{"KEY_RIGHT",             77},
	{"KEY_PLUS_PAD",          78},
	{"KEY_END",               79},
	{"KEY_DOWN",              80},
	{"KEY_PGDN",              81},
	{"KEY_INSERT",            82},
	{"KEY_DEL",               83},
	{"KEY_PRTSCR",            84},
	{"KEY_F11",               87},
	{"KEY_F12",               88},
	{"KEY_LWIN",              91},
	{"KEY_RWIN",              92},
	{"KEY_MENU",              93},
	{"KEY_PAD",              100},
	{"KEY_RCONTROL",         120},
	{"KEY_ALTGR",            121},
	{"KEY_SLASH2",           122},
	{"KEY_PAUSE",            123}
};

static int	macro_number = 0;
static int	macro_pos = 0;
static int	macro_counter;

//-- Global Variables --------------------------------------------------------
int	macro_playing = 0;
				
//----------------------------------------------------------------------------
unsigned char	read_player1(void)
{
	unsigned char	k = 0xFF;
	int		i = 0;

	if ( !macro_playing )
	{
		if (key[controller1_up])
			k &= 0xFE;

		if (key[controller1_down])
			k &= 0xFD;

		if (key[controller1_left])
			k &= 0xFB;

		if (key[controller1_right])
			k &= 0xF7;

		if (key[controller1_a])
			k &= 0xEF;

		if (key[controller1_b])
			k &= 0xDF;

		if (key[controller1_c])
			k &= 0xBF;

		if (key[controller1_d])
			k &= 0x7F;
	
		if (k==0xFF)
		{
			while(macro_list[i].scancode != 0)
			{
				if (key[macro_list[i].scancode] != 0)
				{
					macro_playing = 1;
					macro_number = i;
					macro_pos = 0;
					macro_counter = 5;
				}
			
				i++;
			}
		}
	}

	if ( macro_playing )
	{
		if (macro_list[macro_number].macro[macro_pos] == 0xFF)
			macro_playing = 0;
		else
		{
			k = ~macro_list[macro_number].macro[macro_pos];
			if (macro_counter != 0)
				macro_counter--;
			else
			{
				macro_counter = 5;
				macro_pos++;
			}
		}
	}

	return	k;
}

//----------------------------------------------------------------------------
unsigned char	read_player2(void)
{
	unsigned char	k = 0xFF;

	if (key[controller2_up])
		k &= 0xFE;

	if (key[controller2_down])
		k &= 0xFD;

	if (key[controller2_left])
		k &= 0xFB;

	if (key[controller2_right])
		k &= 0xF7;

	if (key[controller2_a])
		k &= 0xEF;

	if (key[controller2_b])
		k &= 0xDF;

	if (key[controller2_c])
		k &= 0xBF;

	if (key[controller2_d])
		k &= 0x7F;
	
	return	k;
}

//----------------------------------------------------------------------------
char	read_pl12_startsel(void)
{
	char	k = 0x0F;
	
	if (key[controller1_start])
		k &= 0xFE;
		
	if (key[controller1_select])
		k &= 0xFD;

	if (key[controller2_start])
		k &= 0xFB;
		
	if (key[controller2_select])
		k &= 0xF7;

	return k;
}

//----------------------------------------------------------------------------
unsigned int	input_identify(unsigned char *key_name)
{
	unsigned int	i;
	
	for(i=0;i<95;i++)
		if (strcmp((char *)key_name, (char *)keydefs[i].key_name) == 0)
			return keydefs[i].key_scancode;
	
	return 0;
}

//----------------------------------------------------------------------------
unsigned char	*input_whois(unsigned int key_scancode)
{
	unsigned int	i;
	
	for(i=0;i<95;i++)
		if (keydefs[i].key_scancode == key_scancode)
			return keydefs[i].key_name;
	
	return	NULL;
}
