/**************************************
****  NEOCD.C  - Main Source File  ****
**************************************/

//-- Include Files -----------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <allegro.h>
#include "cdaudio/cdaudio.h"
#include "cdrom/cdrom.h"
#include "mc68000/68kregs.h"
#include "shell/shell.h"
#include "memory/memory.h"
#include "video/video.h"
#include "config/config.h"
#include "input/macro.h"

#ifdef	DEBUG
#include "debug/debug.h"
#endif

// Cut off Allegro size...
DECLARE_DIGI_DRIVER_LIST()
DECLARE_MIDI_DRIVER_LIST()
BEGIN_GFX_DRIVER_LIST
	GFX_DRIVER_VESA3
	GFX_DRIVER_VESA2L
	GFX_DRIVER_VESA2B
	GFX_DRIVER_VESA1
END_GFX_DRIVER_LIST


// 12MHz/224lines/60fps = 893 cycles per line
#define	CYCLES_PER_LINE	893

// Version, date & time to display on startup
#define VERSION "NEOGEO-CD Emulator v0.40 ("__DATE__" "__TIME__")"

//-- Global Variables --------------------------------------------------------
char			*neogeo_rom_memory = NULL;
char			*neogeo_prg_memory = NULL;
char			*neogeo_fix_memory = NULL;
char			*neogeo_spr_memory = NULL;
char			global_error[80];
int				neogeo_vline;
int				neogeo_68k_cycles;
int				neogeo_last_command;
unsigned char	neogeo_memorycard[8192];
int				neogeo_prio_mode = 0;

//-- Debugger related stuff --------------------------------------------------
#ifdef	DEBUG
int				neogeo_cmd_repeat_count;
int				neogeo_last_addr;
int				neogeo_last_size;
int				neogeo_last_vaddr;
int				neogeo_last_vsize;
#endif

//-- 68K Core related stuff --------------------------------------------------
int					mame_debug = 0;
char				*OP_ROM;
char				*OP_RAM;
int					previouspc = 0;
int					ophw = 0;
int					cur_mrhard = 0;
extern int			MC68000_ICount;
extern Registers68K	regs;
extern void			M68KRESET(void);	// 68K core init, not reset
extern int			M68KRUN(int);
extern void			initialize_memmap(void);

//-- Function Prototypes -----------------------------------------------------
void	neogeo_init(void);
void	neogeo_reset(void);
void	neogeo_shutdown(void);
void	MC68000_Cause_Interrupt(int);
void	neogeo_exception(void);
void	neogeo_run(void);

//----------------------------------------------------------------------------
int	main(void)
{
	MACROFILE	fm;
	FILE		*fp;
	int			result;
	int			a, b;
	
	atexit(neogeo_shutdown);

	puts(VERSION);
	
	#ifdef	DEBUG
	puts("****DEBUG BUILD****");
	#endif

	puts("NEOGEO: Allocating memory...");

	printf("PRG (2Mo) ... ");
	neogeo_prg_memory = calloc(1, 2097152);
	if (neogeo_prg_memory==NULL) {
		puts("failed !");
		return	0;
	}
	puts("OK!");

	printf("SPR (4Mo) ... ");
	neogeo_spr_memory = calloc(1, 4194304);
	if (neogeo_spr_memory==NULL) {
		puts("failed !");
		return	0;
	}
	puts("OK!");
	
	printf("ROM (128k) ... ");
	neogeo_rom_memory = calloc(1, 131072);
	if (neogeo_rom_memory==NULL) {
		puts("failed !");
		return	0;
	}
	puts("OK!");

	printf("FIX (128k) ... ");
	neogeo_fix_memory = calloc(1, 131072);
	if (neogeo_fix_memory==NULL) {
		puts("failed !");
		return	0;
	}
	puts("OK!");
	
	// Initialize Memory Mapping
	initialize_memmap();
	
	puts("NEOGEO: Loading BIOS...");

	fp = fopen("neocd.bin", "rb");
	if (fp==NULL) {
		puts("NEOGEO: FATAL - Could not load NEOCD.BIN");
		return 0;
	}
	fread(neogeo_rom_memory, 1, 131072, fp);
	fclose(fp);
	swab(neogeo_rom_memory, neogeo_rom_memory, 131072);

	/*** Patch BIOS exit ***/
	*((short*)(neogeo_rom_memory+0x55e)) = 0xFABE;
	*((short*)(neogeo_rom_memory+0x560)) = 0x4E75;
	/*** Patch BIOS load files w/ now loading message ***/
	*((short*)(neogeo_rom_memory+0x552)) = 0xFABF;
	*((short*)(neogeo_rom_memory+0x554)) = 0x4E75;
	/*** Patch BIOS load files w/out now loading ***/
	*((short*)(neogeo_rom_memory+0x564)) = 0xFAC0;
	*((short*)(neogeo_rom_memory+0x566)) = 0x4E75;
	/*** Patch BIOS CDROM Check ***/
	*((short*)(neogeo_rom_memory+0xB040)) = 0x4E71;
	*((short*)(neogeo_rom_memory+0xB042)) = 0x4E71;
	/*** Patch BIOS upload command ***/
	*((short*)(neogeo_rom_memory+0x546)) = 0xFAC1;
	*((short*)(neogeo_rom_memory+0x548)) = 0x4E75;

	/*** Full reset, please ***/
	*((short*)(neogeo_rom_memory+0xA87A)) = 0x4239;
	*((short*)(neogeo_rom_memory+0xA87C)) = 0x0010;
	*((short*)(neogeo_rom_memory+0xA87E)) = 0xFDAE;

	/*** Trap exceptions ***/
	*((short*)(neogeo_rom_memory+0xA5B6)) = 0x4AFC;

	puts("NEOGEO: Loading Startup info...");

	fp = fopen("startup.bin", "rb");
	if (fp==NULL) {
		puts("NEOGEO: FATAL - Could not load STARTUP.BIN");
		return 0;
	}

	fread(neogeo_prg_memory + 0x10F300, 1, 3328, fp);
	fclose(fp);
	swab(neogeo_prg_memory + 0x10F300, neogeo_prg_memory + 0x10F300, 3328);

	fp = fopen("memcard.bin", "rb");
	if (fp!=NULL) {
		puts("NEOGEO: Loading memory card...");
		fread(neogeo_memorycard, 1, 8192, fp);
		fclose(fp);
	}

	// Read config files
	config_read();

	// Initialize Allegro
	allegro_init();

	// Initialize Macros
	macro_init();

	// Install PM keyboard handler
	install_keyboard();
	
	// Initialize CD-DA
	cdda_current_drive = -1;
	result = cdda_init();
	puts(global_error);
	if (!result)
		return	0;

	// Video init
	result = video_init();
	puts(global_error);
	if (!result)
		return	0;

	// CD-ROM init
	result = cdrom_init();
	puts(global_error);
	if (!result)
		return	0;

	neogeo_init();
	
	puts("The system is coming up, please wait...");
	
	// Check for drive number override
	if (cdrom_current_drive == -1)
		cdrom_current_drive = cdda_first_drive;
	
	// Process IPL.TXT
	if (!cdrom_process_ipl()) {
		puts(global_error);
		return 0;
	}

	// Read game name (used in macro system)
	config_read_name();
	strtrim(config_game_name, config_game_name);
	printf("Game name: <%s>\n", config_game_name);

	// Special patch
	if (strcmp(config_game_name, "TEST PROGRAM USA") == 0)
	{
		printf("NEOGEO: Loading SAMURAI SPIRITS RPG patch...\n");
		fp = fopen("patch.prg", "rb");
		fread(neogeo_prg_memory + 0x132000, 1, 112, fp);
		fclose(fp);
		swab(neogeo_prg_memory + 0x132000, neogeo_prg_memory + 0x132000, 112);
	}
	
	// At this point, we are ready!!
	puts("NEOGEO: System is Ready!");

	// First time init
	puts("NEOGEO: Master Reset.");
	memset((void *)&regs, 0, sizeof(regs));
	regs.pc = 0xc0a822;
	regs.sr = 0x2700;
	regs.sr_h = 0x27;
	regs.a7 = regs.isp = regs.usp = 0x10F300;
	OP_ROM = neogeo_rom_memory - 0xC00000;
	neogeo_vline = 0;
	neogeo_68k_cycles = CYCLES_PER_LINE;
	cpu_writemem24_dword(0x10F6EE, cpu_readmem24_dword(0x68L)); // $68 *must* be copied at 10F6EE
	cpu_writemem24(0x00013B, 0x01);

	#ifdef	DEBUG
	debug_dump_regs();
	debug_dump_dasm_pc();
	neogeo_cmd_repeat_count = 0;
	neogeo_last_addr=0;
	neogeo_last_size=16;
	neogeo_last_vaddr=0xE000;
	neogeo_last_vsize=16;
	#endif
	
	neogeo_last_command = -1;//SHELL_GO;
//	neogeo_cmd_repeat_count = 1;

	while(1) {
		#ifdef	DEBUG
		if (!neogeo_cmd_repeat_count) {
		#endif
			shell_get_cmd();
			if (shell_argc!=0)
				neogeo_last_command = shell_identify_command(&shell_argv[0][0]);
		#ifdef	DEBUG
		}
		#endif
	
		switch( neogeo_last_command ) {
		case	SHELL_QUIT :
			return 0;
		#ifdef	DEBUG
		case	SHELL_STEP :
			if (shell_argc>1) {
				sscanf(&shell_argv[1][0], "%d", &neogeo_cmd_repeat_count);
				shell_argc = 1;
			}
			MC68000_ICount = 0;
			M68KRUN(0);
			neogeo_68k_cycles += MC68000_ICount;
			if (neogeo_cmd_repeat_count)
				neogeo_cmd_repeat_count--;

			if (!neogeo_cmd_repeat_count)
				debug_dump_dasm_pc();
			break;
		case	SHELL_REGS :
			debug_dump_regs();
			break;
		case	SHELL_NEXT :
			break;
		case	SHELL_DUMP :
			if (shell_argc>1)
				sscanf(&shell_argv[1][0], "%x", &neogeo_last_addr);
			if (shell_argc>2)
				sscanf(&shell_argv[2][0], "%d", &neogeo_last_size);
			
			neogeo_last_addr = debug_dump_mem(neogeo_last_addr, neogeo_last_size);
			break;
		case	SHELL_VDMP :
			if (shell_argc>1)
				sscanf(&shell_argv[1][0], "%x", &neogeo_last_vaddr);
			if (shell_argc>2)
				sscanf(&shell_argv[2][0], "%d", &neogeo_last_vsize);
			
			neogeo_last_vaddr = debug_dump_vidmem(neogeo_last_vaddr, neogeo_last_vsize);
			break;
		case	SHELL_POKE :
			if (shell_argc>2) {
				sscanf(&shell_argv[1][0], "%x", &a);
				sscanf(&shell_argv[2][0], "%x", &b);
				cpu_writemem24(a, b);
			} else {
				puts("NEOGEO: Too few arguments !");
			}
			break;
		case	SHELL_VSCR :
			video_set_gfx_mode();
			if (neogeo_prio_mode)
				video_draw_screen2();
			else
				video_draw_screen1();
			readkey();
			video_set_txt_mode();
			break;
		case	SHELL_SMEM:
			if (shell_argc<4)
				puts("NEOGEO: Missing argument.");
			else {
				sscanf(&shell_argv[2][0], "%x", &a);
				sscanf(&shell_argv[3][0], "%x", &b);
				fp = fopen(&shell_argv[1][0], "wb");
				if (fp!=NULL) {
					fwrite(neogeo_prg_memory+a, b, 1, fp);
					fclose(fp);
				}
				printf("NEOGEO: Memory saved as %s\n", &shell_argv[1][0]);
			}
			break;
		#endif
		case	SHELL_RSET :
			neogeo_reset();
			#ifdef	DEBUG
			debug_dump_regs();
			debug_dump_dasm_pc();
			#endif
			break;
		case	SHELL_GO :
			neogeo_run();
			break;
		case	SHELL_JAP :
			cpu_writemem24(0x10FD83, 0);
			neogeo_reset();
			break;
		case	SHELL_US :
			cpu_writemem24(0x10FD83, 1);
			neogeo_reset();
			break;
		case	SHELL_SPA :
			cpu_writemem24(0x10FD83, 2);
			neogeo_reset();
			break;
		case	SHELL_MLST :
			if (macro_fopen(&fm, "macro.cfg")) {
				if (macro_find_section(&fm, config_game_name))
				{
					printf("Available macro sets for %s\n", config_game_name);
					macro_print_list(&fm);
				}
				else
					printf("No macro sets available !\n");
					
				macro_fclose(&fm);
			}
			else
				puts("macro.cfg not found !");
			break;
		case	SHELL_MLOD :
			if (shell_argc==2) {
				if (macro_fopen(&fm, "macro.cfg")) {
					if (macro_find_section(&fm, config_game_name))
					{
						if (macro_load(&fm, &shell_argv[1][0])) {
							puts("Loaded macros :");
							macro_avail();
						} else {
							puts("Macro set not found !");
						}
					}
					else
						printf("No macro sets available !\n");

					macro_fclose(&fm);
				} else {
					puts("macro.cfg not found !");
				}
			}
			else
				puts("Macro set name needed !");
			break;
		case	SHELL_MAVL :
			if (macro_list[0].scancode != 0) {
				puts("Available macros :");
				macro_avail();
			} else
				puts("No macro loaded !");
			break;
		
		default :
			puts("NEOGEO: Unknown command.");
		}
		
		#ifdef	DEBUG
		if (neogeo_68k_cycles < 0) {
			neogeo_vline++;
			neogeo_68k_cycles += CYCLES_PER_LINE;
			if (neogeo_vline==224) {
				neogeo_vline=0;
				MC68000_Cause_Interrupt(2);
			}
		}
		#endif
	}
	return 0;
}

void	neogeo_init(void)
{
	shell_init();
	M68KRESET();
	puts("NEOGEO: 68K Emulation Core Started.");
}

void	neogeo_reset(void)
{
	puts("NEOGEO: System Reset.");

	memset((void *)&regs, 0, sizeof(regs));
	regs.pc = 0x122;
	regs.sr = 0x2700;
	regs.sr_h = 0x27;
	regs.a7 = regs.isp = regs.usp = 0x10F300;
	OP_ROM = neogeo_prg_memory;
	neogeo_vline = 0;
	neogeo_68k_cycles = CYCLES_PER_LINE;
	cpu_writemem24(0x10FD80, 0x82);
	cpu_writemem24(0x10FDAF, 0x01);
	cpu_writemem24(0x10FEE1, 0x0A);
	cpu_writemem24(0x10F675, 0x01);
	cpu_writemem24(0x10FEBF, 0x00);
	cpu_writemem24_dword(0x10FDB6, 0);
	cpu_writemem24_dword(0x10FDBA, 0);
}

void	neogeo_shutdown(void)
{
	FILE	*fp;
	
//	video_set_txt_mode();
	puts("NEOGEO: System Shutdown.");
	remove_keyboard();
	cdda_shutdown();
	cdrom_shutdown();
	video_shutdown();

	fp = fopen("memcard.bin", "wb");
	fwrite(neogeo_memorycard, 1, 8192, fp);
	fclose(fp);

	free(neogeo_prg_memory);
	free(neogeo_rom_memory);
	free(neogeo_spr_memory);
	free(neogeo_fix_memory);
}

void	neogeo_exception(void)
{
	video_set_txt_mode();
	printf("NEOGEO: Exception Trapped at %08x !\n", previouspc);
	#ifdef	DEBUG
	neogeo_cmd_repeat_count = 0;
	debug_dump_regs();
	debug_dump_dasm_pc();
	#endif
	readkey();
	video_set_gfx_mode();
}	

void MC68000_Cause_Interrupt(int level)
{
	if (level >= 1 && level <= 7)
        regs.irq |= 1 << (level-1);
}

void	neogeo_exit(void)
{
	video_set_txt_mode();
	puts("NEOGEO: Exit requested by software...");
	exit(0);
}

void	neogeo_run(void)
{
	video_set_gfx_mode();
		
	while(!key[KEY_TILDE])
	{
		MC68000_ICount = 0;
		M68KRUN(CYCLES_PER_LINE);

		neogeo_68k_cycles += MC68000_ICount;
		if (neogeo_68k_cycles < 0) {
			neogeo_vline++;
			neogeo_68k_cycles += CYCLES_PER_LINE;
			if (neogeo_vline==224) {
				neogeo_vline=0;

				if (key[KEY_F11])
				{
					video_hide_fps ^= 1;
					while(key[KEY_F11]);
				}

				if (neogeo_prio_mode)
					video_draw_screen2();
				else
					video_draw_screen1();

				if (key[KEY_F12])
					video_save_snapshot();

				MC68000_Cause_Interrupt(2);
			}
		}
	}

	video_set_txt_mode();
}


// This is a really dirty hack to make SAMURAI SPIRITS RPG work

void	neogeo_prio_switch(void)
{
//	video_set_txt_mode();
//	debug_dump_regs();
//	clear_keybuf();
//	readkey();
//	video_set_gfx_mode();
	
	if (regs.d7 == 0xFFFF)
		return;
	
	if ((regs.d7==9)&&(regs.a3==0x10DED9)&&((regs.a2==0x1081d0)||((regs.a2&0xFFF000)==0x102000)))
	{
		neogeo_prio_mode = 0;
		return;
	}
	
	if ((regs.d7==8)&&(regs.a3==0x10DEC7)&&(regs.a2==0x102900))
	{
		neogeo_prio_mode = 0;
		return;
	}
	
	if (regs.a7 == 0x10F29C)
	{
		if ((regs.d4&0x4010)==0x4010)
		{
			neogeo_prio_mode = 0;
			return;
		}
		
		neogeo_prio_mode = 1;
	}
	else
	{
		if (regs.a3==0x5140)
		{
			neogeo_prio_mode = 1;
			return;
		}

		if ( (regs.a3&~0xF) == (regs.a4&~0xF) )
			neogeo_prio_mode = 1;
		else
			neogeo_prio_mode = 0;
	}
}
